from .exceptions import CBRRequestException
import requests
from datetime import date
import xml.etree.ElementTree as ET
from moneybox.settings import CBR_TIMEOUT, CBR_URL


class CBRClient:
    def __init__(self, url: str = CBR_URL, timeout: int = CBR_TIMEOUT) -> None:
        self.url = url
        self.timeout = timeout

    def get_currencies_rates(self, target_date=None):
        """Target_date must be str in dd/mm/YYYY format. Example 01/01/2000."""

        currencies = {}

        if target_date is None:
            target_date = date.today().strftime("%d/%m/%Y")

        try:
            response = requests.get(url=self.url + "?date_req=" + target_date, timeout=self.timeout)
        except requests.exceptions.RequestException as e:
            raise CBRRequestException(e)

        root = ET.fromstring(response.text)
        for currency in root:
            valute_name = currency.find("CharCode").text
            valute_data = {
                "num_code": currency.find("NumCode").text,
                "char_code": currency.find("CharCode").text,
                "nominal": float(currency.find("Nominal").text),
                "name": currency.find("Name").text,
                "value": float(currency.find("Value").text.replace(",", ".")),
                "cbr_currency_id": currency.attrib["ID"],
            }
            currencies.setdefault(valute_name, valute_data)
        return currencies


cbr_client = CBRClient()



from rest_framework.serializers import ModelSerializer

from wallet.models.group import Group


class GroupSerializer(ModelSerializer):
    class Meta:
        model = Group
        fields = "__all__"



from rest_framework import serializers

from users.models import APIUser
from wallet.models.invite import Invite


class APIUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = APIUser
        fields = ("token",)


class SignupSerializer(serializers.Serializer):
    """Create token and add user to the group."""

    invite_code = serializers.IntegerField(required=False)

    def validate(self, data):
        invite_code = data.get("invite_code")
        if not invite_code:
            return data
        group_invite = Invite.objects.filter(invite_code=invite_code).first()
        if not group_invite or group_invite.is_expired:
            raise serializers.ValidationError("Invalid invite code")
        return data



REPORT_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
REPORT_STANDART_DATE_FORMAT = "%d-%m-%Y"
RATE_DATE_FORMAT = "%YYYY-MM-DD"


def convert_date_to_datetime_format(date):
    return date.strftime(REPORT_DATETIME_FORMAT)


def convert_date_to_standart_format(date):
    return date.strftime(REPORT_STANDART_DATE_FORMAT)



from rest_framework.serializers import ModelSerializer

from wallet.models.transfer import Transfer


class TransferSerializer(ModelSerializer):
    class Meta:
        model = Transfer
        fields = "__all__"



# coding: utf-8

"""
    MoneyBox

    MoneyBox will help you better understand your finances, improve financial literacy, and achieve financial goals.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic import BaseModel, Field, StrictInt, constr, validator

class CurrencyRate(BaseModel):
    """
    CurrencyRate
    """
    id: StrictInt = Field(...)
    currency: Optional[StrictInt] = Field(None, description="Currency")
    rate: constr(strict=True) = Field(..., description="The rate at which the source currencycan be exchanged for the target currency.")
    __properties = ["id", "currency", "rate"]

    @validator('rate')
    def rate_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^-?\d{0,7}(?:\.\d{0,5})?$", value):
            raise ValueError(r"must validate the regular expression /^-?\d{0,7}(?:\.\d{0,5})?$/")
        return value

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> CurrencyRate:
        """Create an instance of CurrencyRate from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                            "id",
                          },
                          exclude_none=True)
        # set to None if currency (nullable) is None
        # and __fields_set__ contains the field
        if self.currency is None and "currency" in self.__fields_set__:
            _dict['currency'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> CurrencyRate:
        """Create an instance of CurrencyRate from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return CurrencyRate.parse_obj(obj)

        _obj = CurrencyRate.parse_obj({
            "id": obj.get("id"),
            "currency": obj.get("currency"),
            "rate": obj.get("rate")
        })
        return _obj





from decimal import Decimal

from django.db import models, transaction

from wallet.models.currency import CurrencyRate
from wallet.models.group import Group
from wallet.models.mixins import TimestampMixin, SafeDeletionMixin, SafeDeletionManager
from wallet.models.wallet import Wallet
from users.models import APIUser


class Transfer(TimestampMixin, SafeDeletionMixin):
    from_wallet = models.ForeignKey(
        Wallet,
        related_name="transfers_from",
        on_delete=models.CASCADE,
        verbose_name="From Wallet",
        help_text="The wallet from which the transfer is made.",
        db_index=True,
    )
    to_wallet = models.ForeignKey(
        Wallet,
        related_name="transfers_to",
        on_delete=models.CASCADE,
        verbose_name="To Wallet",
        help_text="The wallet to which the transfer is made.",
        db_index=True,
    )
    amount = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        verbose_name="Transfer Amount",
        help_text="The amount of money transferred.",
    )
    comment = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name="Transfer Comment",
        help_text="Additional comment about the transfer (optional).",
    )
    created_by = models.ForeignKey(
        APIUser,
        on_delete=models.CASCADE,
        verbose_name="Created by",
        help_text="The user who made the transfer.",
    )
    group = models.ForeignKey(
        Group,
        on_delete=models.CASCADE,
        verbose_name="Group",
        help_text="The group to which the transfer belongs.",
        db_index=True,
    )
    objects = SafeDeletionManager()

    class Meta:
        verbose_name = "Transfer"
        verbose_name_plural = "Transfers"

    @transaction.atomic
    def save(self, *args, **kwargs):
        # TODO add checking of current and target currencies
        super().save(*args, **kwargs)
        rate_from = CurrencyRate.objects.get(
            source_currency=self.from_wallet.currency,
            target_currency=self.to_wallet.currency,
        )
        amount_converted = Decimal(self.amount) * rate_from.rate
        self.from_wallet.balance -= self.amount
        self.from_wallet.save()
        self.to_wallet.balance += amount_converted
        self.to_wallet.save()
        self.from_wallet.balance -= self.amount
        self.from_wallet.save()
        self.to_wallet.balance += amount_converted
        self.to_wallet.save()



from rest_framework import serializers
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field


class ReportSerializer(serializers.Serializer):
    balance = serializers.SerializerMethodField()
    total_incomes = serializers.SerializerMethodField()
    total_expenses = serializers.SerializerMethodField()
    income_expense_ratio = serializers.SerializerMethodField()
    category_incomes = serializers.SerializerMethodField()
    category_expenses = serializers.SerializerMethodField()

    @staticmethod
    @extend_schema_field(OpenApiTypes.DECIMAL)
    def get_balance(obj):
        return obj.get("balance")

    @staticmethod
    @extend_schema_field(OpenApiTypes.DECIMAL)
    def get_total_incomes(obj):
        return obj.get("total_incomes")

    @staticmethod
    @extend_schema_field(OpenApiTypes.DECIMAL)
    def get_total_expenses(obj):
        return obj.get("total_expenses")

    @staticmethod
    @extend_schema_field(OpenApiTypes.DECIMAL)
    def get_income_expense_ratio(obj):
        return obj.get("income_expense_ratio")

    @staticmethod
    @extend_schema_field(OpenApiTypes.OBJECT)
    def get_category_incomes(obj):
        return obj.get("category_incomes")

    @staticmethod
    @extend_schema_field(OpenApiTypes.OBJECT)
    def get_category_expenses(obj):
        return obj.get("category_expenses")



from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication

from api.encryption import encrypt_token
from moneybox.settings import AUTH_HEADER
from users.models import APIUser


class APIAuthentication(BaseAuthentication):
    def authenticate(self, request):
        auth_header = request.headers.get(AUTH_HEADER)
        if not auth_header:
            return None
        header_content = auth_header.split(" ")
        if len(header_content) != 2:
            raise exceptions.AuthenticationFailed("Authorization header should contain 'Token <token_value>'.")
        token_header = header_content[1]
        token_db = encrypt_token(token_header.encode())
        if not APIUser.objects.filter(token=token_db).exists():
            raise exceptions.AuthenticationFailed("No such token found.")
        authenticated_user = APIUser.objects.filter(token=token_db).first()
        return authenticated_user, authenticated_user.token



from rest_framework.serializers import ModelSerializer

from wallet.models.wallet import Wallet


class WalletSerializer(ModelSerializer):
    class Meta:
        model = Wallet
        fields = "__all__"



from django.contrib import admin

from wallet.models.group import Group


@admin.register(Group)
class GroupAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "created_at",
        "updated_at",
    )
    list_filter = (
        "created_at",
        "updated_at",
    )
    raw_id_fields = ("members",)
    date_hierarchy = "created_at"



from django.contrib import admin

from wallet.models.transfer import Transfer


@admin.register(Transfer)
class TransferAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "from_wallet",
        "to_wallet",
        "amount",
        "comment",
        "group",
        "created_at",
        "updated_at",
    )
    list_filter = (
        "created_at",
        "updated_at",
    )
    date_hierarchy = "created_at"



from django.contrib import admin
from wallet.models.currency import Currency, CurrencyRate


@admin.register(Currency)
class CurrencyAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "code",
        "name",
        "type",
        "created_at",
        "updated_at",
    )
    list_filter = ("type", "created_at", "updated_at")
    search_fields = ("name", "code")
    date_hierarchy = "updated_at"


@admin.register(CurrencyRate)
class CurrencyRateAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "currency",
        "rate",
        "created_at",
        "updated_at",
    )
    list_filter = ("currency__type", "created_at", "updated_at")
    date_hierarchy = "created_at"



from pyecharts import options as opts
from pyecharts.charts import Bar, Pie
from pyecharts.globals import ThemeType

from core.defs.exeptions import ReportAPIException


def render_bar_chart(x_axis_data, report_data):
    try:
        bar = (
            Bar(init_opts=opts.InitOpts(theme=ThemeType.LIGHT))
            .add_xaxis(x_axis_data)
            .add_yaxis("Расходы", [report_data["total_expenses"]])
            .add_yaxis("Доходы", [report_data["total_incomes"]])
            .set_global_opts(
                title_opts=opts.TitleOpts(title="Диаграмма доходов и расходов", subtitle="За выбранный период")
            )
        )
        return bar.render_embed()
    except Exception as e:
        raise ReportAPIException(detail=f"Error occurred while rendering bar chart: {e}")


def render_pie_chart(data, title):
    try:
        pie = (
            Pie()
            .add("", data)
            .set_global_opts(title_opts=opts.TitleOpts(title=title))
            .set_series_opts(label_opts=opts.LabelOpts(formatter="{b}: {d}%"))
        )
        return pie.render_embed()
    except Exception as e:
        raise ReportAPIException(detail=f"Error occurred while rendering pie chart: {e}")


def render_charts_to_html(*args):
    try:
        chart_html = " ".join(args)
        return chart_html
    except Exception as e:
        raise ReportAPIException(detail=f"Error occurred while rendering charts to HTML: {e}")


def render_no_data_html():
    return """
    <!DOCTYPE html>
    <html>
    <head>
        <title>No Data to Build a Report</title>
    </head>
    <body>
        <p>No data to build a report</p>
    </body>
    </html>
    """


def generate_charts(x_axis_data, report_data):
    try:
        if (
            not report_data
            or all(value == 0 for value in report_data.values())
            or all(not value for value in report_data.values())
        ):
            return render_no_data_html()
        bar_chart_html = render_bar_chart(x_axis_data, report_data)
        category_incomes = report_data["category_incomes"]
        category_expenses = report_data["category_expenses"]

        incomes_data = [(category["category__name"], category["category_incomes"]) for category in category_incomes]
        pie_chart_incomes_html = render_pie_chart(incomes_data, "Доходы по категориям")

        expenses_data = [(category["category__name"], category["total_expenses"]) for category in category_expenses]
        pie_chart_expenses_html = render_pie_chart(expenses_data, "Расходы по категориям")

        chart_html = render_charts_to_html(bar_chart_html, pie_chart_incomes_html, pie_chart_expenses_html)
        return chart_html
    except Exception as e:
        raise ReportAPIException(detail=f"Error occurred while generating charts: {e}")



from drf_spectacular.extensions import OpenApiAuthenticationExtension


class AuthenticationScheme(OpenApiAuthenticationExtension):
    target_class = "api.authentication.APIAuthentication"
    name = "APIAuthentication"

    def get_security_definition(self, auto_schema):
        return {
            "type": "apiKey",
            "in": "header",
            "name": "Authorization",
        }



from wallet.models.currency import CryptoCurrency


class CoinGeckoCrypto:
    BITCOIN = "bitcoin"
    TON = "the-open-network"
    ETHEREUM = "ethereum"
    USDC = "usd-coin"

    @classmethod
    def map_main_crypto_to_coingecko(cls):
        return {
            CryptoCurrency.TON: cls.TON,
            CryptoCurrency.BITCOIN: cls.BITCOIN,
            CryptoCurrency.ETHEREUM: cls.ETHEREUM,
            CryptoCurrency.USDC: cls.USDC,
        }

    @classmethod
    def map_coingecko_to_main_crypto(cls, key):
        return {
            cls.TON: CryptoCurrency.TON,
            cls.BITCOIN: CryptoCurrency.BITCOIN,
            cls.ETHEREUM: CryptoCurrency.ETHEREUM,
            cls.USDC: CryptoCurrency.USDC,
        }.get(key)



from django.contrib import admin

from wallet.models.income import Income, IncomeCategory


@admin.register(IncomeCategory)
class IncomeCategoryAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "group",
        "created_at",
        "updated_at",
    )
    list_filter = (
        "created_at",
        "updated_at",
    )
    search_fields = ("name",)
    date_hierarchy = "created_at"


@admin.register(Income)
class IncomeAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "amount",
        "category",
        "comment",
        "created_by",
        "wallet",
        "group",
        "created_at",
        "updated_at",
    )
    list_filter = (
        "created_at",
        "updated_at",
    )
    date_hierarchy = "created_at"



import requests

from clients.currency.coingecko.defs import CoinGeckoCrypto
from clients.currency.coingecko.exceptions import CoinGeckoRequestException
from moneybox.settings import COINGECKO_URL, COINGECKO_TIMEOUT
from wallet.models.currency import FiatCurrency


class CoinGeckoClient:
    def __init__(self, url: str = COINGECKO_URL, timeout: int = COINGECKO_TIMEOUT) -> None:
        self.url = url
        self.timeout = timeout

    def get_rates(self):
        try:
            response = requests.get(
                url=self.url.format(
                    crypto_currencies=",".join(CoinGeckoCrypto.map_main_crypto_to_coingecko().values()),
                    fiat_currency=FiatCurrency.RUB,
                ),
                timeout=self.timeout,
            )
            rates = response.json()
        except requests.exceptions.RequestException as e:
            raise CoinGeckoRequestException(e)
        result = dict()
        for k, v in rates.items():
            result[CoinGeckoCrypto.map_coingecko_to_main_crypto(k)] = round(v.get(FiatCurrency.RUB.lower()), 4)
        return result


coingecko_client = CoinGeckoClient()



import random
from datetime import timedelta

from django.utils import timezone
from drf_spectacular.utils import extend_schema
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from api.permissions import IsAuthenticated
from api.serializers.invite import InviteSerializer
from wallet.models.group import Group
from wallet.models.invite import Invite


@extend_schema(tags=["Invites"])
class InviteViewSet(ModelViewSet):
    queryset = Invite.objects.all()
    serializer_class = InviteSerializer

    @action(detail=False, methods=("POST",), permission_classes=(IsAuthenticated,))
    def invite(self, request):  # TODO scheme for swagger
        """Create invite code."""
        user_token = request.user.token
        group = Group.objects.filter(members__token=user_token).first()
        if group:
            invite_code = random.randint(1000000, 9999999)
            expires_at = timezone.now() + timedelta(days=7)
            Invite.objects.create(invite_code=invite_code, group=group, expires_at=expires_at)
            return Response({"code": invite_code})
        return Response({"detail": "Group not found"})  # TODO unreal case



# flake8: noqa

# import apis into api package
from openapi_client.api.auth_api import AuthApi
from openapi_client.api.currency_api import CurrencyApi
from openapi_client.api.expenses_api import ExpensesApi
from openapi_client.api.incomes_api import IncomesApi
from openapi_client.api.invites_api import InvitesApi
from openapi_client.api.reports_api import ReportsApi
from openapi_client.api.transfers_api import TransfersApi
from openapi_client.api.users_api import UsersApi
from openapi_client.api.wallets_api import WalletsApi
from openapi_client.api.group_api import GroupApi




# Generated by Django 3.2 on 2023-10-03 11:41

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('users', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Currency',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('code', models.CharField(help_text='The unique code for this currency', max_length=10, unique=True, verbose_name='Currency Code')),
                ('name', models.CharField(help_text='The name of the currency, e.g. US Dollar', max_length=255, verbose_name='Currency Name')),
                ('type', models.CharField(choices=[('fiat', 'fiat'), ('crypto', 'crypto')], default='fiat', help_text='The type of the currency, e.g. fiat or crypto', max_length=255, verbose_name='Currency Type')),
            ],
            options={
                'verbose_name': 'Currency',
                'verbose_name_plural': 'Currencies',
            },
        ),
        migrations.CreateModel(
            name='Group',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('members', models.ManyToManyField(db_index=True, help_text='Members of the group', related_name='groups', to='users.APIUser', verbose_name='Group members')),
            ],
            options={
                'verbose_name': 'Group',
                'verbose_name_plural': 'Groups',
            },
        ),
        migrations.CreateModel(
            name='Wallet',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('name', models.CharField(db_index=True, help_text='Name of the wallet', max_length=255, verbose_name='Name')),
                ('balance', models.DecimalField(decimal_places=2, help_text='Current balance of the wallet', max_digits=10, verbose_name='Balance')),
                ('created_by', models.ForeignKey(help_text='Owner of the wallet', on_delete=django.db.models.deletion.CASCADE, to='users.apiuser', verbose_name='Created by')),
                ('currency', models.ForeignKey(help_text='Currency of the wallet', on_delete=django.db.models.deletion.CASCADE, to='wallet.currency', verbose_name='Currency')),
                ('group', models.ForeignKey(help_text='Group that the wallet belongs to', on_delete=django.db.models.deletion.CASCADE, to='wallet.group', verbose_name='Group')),
            ],
            options={
                'verbose_name': 'Wallet',
                'verbose_name_plural': 'Wallets',
            },
        ),
        migrations.CreateModel(
            name='Transfer',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('amount', models.DecimalField(decimal_places=2, help_text='The amount of money transferred.', max_digits=10, verbose_name='Transfer Amount')),
                ('comment', models.CharField(blank=True, help_text='Additional comment about the transfer (optional).', max_length=255, null=True, verbose_name='Transfer Comment')),
                ('created_by', models.ForeignKey(help_text='The user who made the transfer.', on_delete=django.db.models.deletion.CASCADE, to='users.apiuser', verbose_name='Created by')),
                ('from_wallet', models.ForeignKey(help_text='The wallet from which the transfer is made.', on_delete=django.db.models.deletion.CASCADE, related_name='transfers_from', to='wallet.wallet', verbose_name='From Wallet')),
                ('group', models.ForeignKey(help_text='The group to which the transfer belongs.', on_delete=django.db.models.deletion.CASCADE, to='wallet.group', verbose_name='Group')),
                ('to_wallet', models.ForeignKey(help_text='The wallet to which the transfer is made.', on_delete=django.db.models.deletion.CASCADE, related_name='transfers_to', to='wallet.wallet', verbose_name='To Wallet')),
            ],
            options={
                'verbose_name': 'Transfer',
                'verbose_name_plural': 'Transfers',
            },
        ),
        migrations.CreateModel(
            name='Invite',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('invite_code', models.IntegerField()),
                ('expires_at', models.DateTimeField()),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='invites', to='wallet.group')),
            ],
            options={
                'verbose_name': 'Invite',
                'verbose_name_plural': 'Invites',
            },
        ),
        migrations.CreateModel(
            name='IncomeCategory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('name', models.CharField(db_index=True, help_text='Name of the income category', max_length=255, verbose_name='Name')),
                ('created_by', models.ForeignKey(help_text='Profile that created the income category', on_delete=django.db.models.deletion.CASCADE, related_name='income_categories', to='users.apiuser', verbose_name='Created by')),
                ('group', models.ForeignKey(help_text='Group that the income category belongs to', on_delete=django.db.models.deletion.CASCADE, to='wallet.group', verbose_name='Group')),
            ],
            options={
                'verbose_name': 'Income Category',
                'verbose_name_plural': 'Income Categories',
            },
        ),
        migrations.CreateModel(
            name='Income',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=10, verbose_name='Income amount')),
                ('comment', models.CharField(blank=True, max_length=255, null=True, verbose_name='Comment on income')),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wallet.incomecategory', verbose_name='Income category')),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='users.apiuser', verbose_name='Income creator')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wallet.group', verbose_name='Group for income')),
                ('wallet', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wallet.wallet', verbose_name='Wallet for income')),
            ],
            options={
                'verbose_name': 'Income',
                'verbose_name_plural': 'Incomes',
            },
        ),
        migrations.CreateModel(
            name='ExpenseCategory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('name', models.CharField(db_index=True, help_text='The name of the expense category', max_length=255, verbose_name='Name')),
                ('created_by', models.ForeignKey(help_text='The user who created this expense category', on_delete=django.db.models.deletion.CASCADE, to='users.apiuser', verbose_name='Created by')),
                ('group', models.ForeignKey(help_text='The group this expense category belongs to', on_delete=django.db.models.deletion.CASCADE, to='wallet.group', verbose_name='Group')),
            ],
            options={
                'verbose_name': 'Expense Category',
                'verbose_name_plural': 'Expense Categories',
            },
        ),
        migrations.CreateModel(
            name='Expense',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=10, verbose_name='Amount of expense')),
                ('comment', models.CharField(blank=True, max_length=255, null=True, verbose_name='Comment on expense')),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wallet.expensecategory', verbose_name='Expense category')),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='users.apiuser', verbose_name='User who made the expense')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wallet.group', verbose_name='Group related to the expense')),
                ('wallet', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wallet.wallet', verbose_name='Wallet used for the expense')),
            ],
            options={
                'verbose_name': 'Expense',
                'verbose_name_plural': 'Expenses',
            },
        ),
        migrations.CreateModel(
            name='CurrencyRate',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Date and time of creation', verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, db_index=True, help_text='Date and time of last update', verbose_name='Updated at')),
                ('is_deleted', models.BooleanField(default=False)),
                ('rate', models.DecimalField(decimal_places=5, help_text='The rate at which the source currencycan be exchanged for the target currency.', max_digits=12, verbose_name='Exchange Rate')),
                ('currency', models.ForeignKey(blank=True, help_text='Currency', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='rate', to='wallet.currency', verbose_name='Currency')),
            ],
            options={
                'verbose_name': 'Currency rate',
                'verbose_name_plural': 'Currency rates',
                'get_latest_by': ('created_at',),
            },
        ),
        migrations.AddConstraint(
            model_name='invite',
            constraint=models.UniqueConstraint(fields=('invite_code', 'group'), name='unique_invite'),
        ),
    ]



# coding: utf-8

"""
    MoneyBox

    MoneyBox will help you better understand your finances, improve financial literacy, and achieve financial goals.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import List, Optional
from pydantic import BaseModel, Field, StrictBool, StrictInt, conlist

class PatchedGroup(BaseModel):
    """
    PatchedGroup
    """
    id: Optional[StrictInt] = None
    created_at: Optional[datetime] = Field(None, description="Date and time of creation")
    updated_at: Optional[datetime] = Field(None, description="Date and time of last update")
    is_deleted: Optional[StrictBool] = None
    members: Optional[conlist(StrictInt)] = Field(None, description="Members of the group")
    __properties = ["id", "created_at", "updated_at", "is_deleted", "members"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> PatchedGroup:
        """Create an instance of PatchedGroup from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                            "id",
                            "created_at",
                            "updated_at",
                          },
                          exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> PatchedGroup:
        """Create an instance of PatchedGroup from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return PatchedGroup.parse_obj(obj)

        _obj = PatchedGroup.parse_obj({
            "id": obj.get("id"),
            "created_at": obj.get("created_at"),
            "updated_at": obj.get("updated_at"),
            "is_deleted": obj.get("is_deleted"),
            "members": obj.get("members")
        })
        return _obj





from wallet.admin.currency import CurrencyAdmin, CurrencyRateAdmin
from wallet.admin.expense import ExpenseAdmin, ExpenseCategoryAdmin
from wallet.admin.group import GroupAdmin
from wallet.admin.income import IncomeAdmin, IncomeCategoryAdmin
from wallet.admin.transfer import TransferAdmin
from wallet.admin.wallet import WalletAdmin
from wallet.admin.invite import InviteAdmin

__all__ = [
    "CurrencyAdmin",
    "CurrencyRateAdmin",
    "ExpenseAdmin",
    "ExpenseCategoryAdmin",
    "GroupAdmin",
    "IncomeAdmin",
    "IncomeCategoryAdmin",
    "TransferAdmin",
    "WalletAdmin",
    "InviteAdmin",
]



from drf_spectacular.utils import extend_schema
from rest_framework.viewsets import ModelViewSet

from api.serializers import TransferSerializer
from wallet.models.transfer import Transfer


@extend_schema(tags=["Transfers"])
class TransferViewSet(ModelViewSet):
    queryset = Transfer.objects.order_by("pk")
    serializer_class = TransferSerializer



import pytest
from decimal import Decimal

from django.contrib.auth import get_user_model
from django.utils import timezone
from rest_framework.test import APIRequestFactory

from api.serializers.report import ReportSerializer
from api.views.auth import decrypt_ciphertext, encrypt_token
from api.views.report import ReportViewSet
from core.defs.constants import Currency as ConstantsCurrency
from core.defs.datetime import convert_date_to_datetime_format
from users.models import APIUser
from wallet.models.currency import Currency
from wallet.models.expense import Expense, ExpenseCategory
from wallet.models.group import Group
from wallet.models.income import Income, IncomeCategory
from wallet.models.wallet import Wallet


User = get_user_model()


class TestReport:
    @pytest.fixture
    def api_rf(self):
        return APIRequestFactory()

    @pytest.fixture
    def user(self):
        return APIUser.objects.create(token="8db757d5-c7d9-44dc-a7a7-f9cf8e7920dd")

    @pytest.fixture
    def group(self, user):
        group = Group.objects.create(name="Test Group")
        group.members.set([user.id])
        return group

    @pytest.fixture
    def currency(self):
        return Currency.objects.create(*ConstantsCurrency.RUB)

    @pytest.fixture
    def wallet(self, user, group, currency):
        return Wallet.objects.create(name="Test Wallet", balance=50, created_by=user, group=group, currency=currency)

    @pytest.fixture
    def income_category(self, user, group):
        return IncomeCategory.objects.create(name="Test Income Category", group=group, created_by=user)

    @pytest.fixture
    def expense_category(self, user, group):
        return ExpenseCategory.objects.create(name="Test Expense Category", group=group, created_by=user)

    @pytest.fixture
    def income(self, user, income_category, group, wallet):
        income = Income.objects.create(
            amount=100, category=income_category, created_by=user, group=group, wallet=wallet, comment="Зарплата"
        )
        income.save()
        return income

    @pytest.fixture
    def expense(self, user, expense_category, group, wallet):
        expense = Expense.objects.create(
            amount=50, category=expense_category, created_by=user, group=group, wallet=wallet, comment="Такси"
        )
        expense.save()
        return expense

    @pytest.mark.django_db
    def test_list(self, api_rf, user):
        request = api_rf.get("/reports/")
        request.user = user

        # Add a token to the request
        token = user.auth_token
        request.META["HTTP_AUTHORIZATION"] = f"Token {token.key}"

        # Get actual data
        start_date = timezone.now().date()
        end_date = timezone.now().date()
        total_incomes_per, total_incomes = ReportViewSet.get_total_incomes(user, start_date, end_date)
        total_expenses_per, total_expenses = ReportViewSet.get_total_expenses(user, start_date, end_date)
        income_expense_ratio = ReportViewSet.get_income_expense_ratio(total_incomes_per, total_expenses_per)
        category_incomes = ReportViewSet.get_category_incomes(user, start_date, end_date)
        category_expenses = ReportViewSet.get_category_expenses(user, start_date, end_date)

        actual_data = {
            "balance": total_incomes - total_expenses,
            "total_incomes": total_incomes_per,
            "total_expenses": total_expenses_per,
            "income_expense_ratio": income_expense_ratio,
            "category_incomes": category_incomes,
            "category_expenses": category_expenses,
        }

        expected_data = {
            "balance": Decimal("50"),
            "total_incomes": Decimal("100"),
            "total_expenses": Decimal("50"),
            "income_expense_ratio": Decimal("2"),
            "category_incomes": [
                {
                    "category__name": "Test Income Category",
                    "total_expenses": Decimal("100"),
                    "created_at": convert_date_to_datetime_format(timezone.now()),
                }
            ],
            "category_expenses": [
                {
                    "category__name": "Test Expense Category",
                    "total_expenses": Decimal("50"),
                    "created_at": convert_date_to_datetime_format(timezone.now()),
                }
            ],
        }

        response = ReportViewSet.as_view({"get": "list"})(request)

        assert response.status_code == 200

        assert ReportSerializer(actual_data).data == ReportSerializer(expected_data).data


class TestEncryption:
    token = "8db757d5-c7d9-44dc-a7a7-f9cf8e7920dd"
    token_encrypted_pt_1 = b"\x01rf8>\x01\xfe\x8a\xd19\x00\xb9\xcd<W\xef\xa3\x03\xc0\xf5\xf3\\U93\xd9\xfb\xac\xf3\xd0"
    token_encrypted_pt_2 = b"\x90\x1c5\xda\xd0K\xbb\xe6\x810b\xe4\xd8\xf2\xed\xd4\x9b'\xaf\xffi\x1f\\\x07\x86\xfa\x7f"
    token_encrypted = token_encrypted_pt_1 + token_encrypted_pt_2

    def test_encrypt_token(self):
        assert encrypt_token(self.token.encode()) == self.token_encrypted

    def test_decrypt_ciphertext(self):
        assert decrypt_ciphertext(self.token_encrypted) == self.token



class Currency:
    RUB = ("RUB", "Рубль")
    USD = ("USD", "Доллар")
    EUR = ("EUR", "Евро")


DEFAULT_EXPENSE_CATEGORY = [
    "Авто",
    "Благотворительность",
    "Бытовая техника",
    "Дача",
    "Дети",
    "Долги",
    "Домашние животные",
    "ЖКХ",
    "Кредит",
    "Медицина",
    "Образование",
    "Одежда и обувь",
    "Питание",
    "Подарки",
    "Прочее",
    "Развлечения",
    "Ремонт",
    "Связь",
    "Страхование",
    "Транспорт",
]

DEFAULT_INCOME_CATEGORY = [
    "Аванс",
    "Алименты",
    "Банковские проценты",
    "Возврат налогов",
    "Возврат платежа",
    "Дивиденты",
    "Доход от бизнеса",
    "Зарплата",
    "Командировочные",
    "Отпускные",
    "Пенсия",
    "Подарки",
    "Подработка",
    "Помощь",
    "Премия",
    "Приз(выигрыш)",
    "Продажа вещей",
    "Социальное пособие",
    "Стипендия",
]
WALLET_LIST = [
    "Карта",
    "Наличные",
    "Счет",
]

RUB_NAME = "Российский рубль"



from api.permissions import IsAdminOrReadOnly
from drf_spectacular.utils import extend_schema
from rest_framework.viewsets import ModelViewSet

from api.serializers import CurrencySerializer, CurrencyRateSerializer
from wallet.models.currency import Currency, CurrencyRate
from rest_framework import filters
from rest_framework.decorators import action
from django.http import HttpResponse
from datetime import datetime
from core.defs.datetime import RATE_DATE_FORMAT


@extend_schema(tags=["Currency"])
class CurrencyViewSet(ModelViewSet):
    """List of currencies or add new currency"""

    queryset = Currency.objects.order_by("pk")
    serializer_class = CurrencySerializer
    filter_backends = (filters.SearchFilter,)
    search_fields = (
        "code",
        "name",
    )
    permission_classes = (IsAdminOrReadOnly,)


@extend_schema(tags=["Currency"])
class CurrencyRateViewSet(ModelViewSet):
    queryset = CurrencyRate.objects.order_by("pk")
    serializer_class = CurrencyRateSerializer

    @action(
        detail=False,
        methods=["GET"],
    )
    def get_rate(self, request):
        currency_from = request.query_params.get("currency_from")
        currency_to = request.query_params.get("currency_to")
        date = request.query_params.get("date")
        date = datetime.strptime(date, RATE_DATE_FORMAT).date()
        rate = CurrencyRate.get_exchange_rate(currency_from, currency_to, date)
        return HttpResponse(rate)



# coding: utf-8

"""
    MoneyBox

    MoneyBox will help you better understand your finances, improve financial literacy, and achieve financial goals.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from openapi_client.api.incomes_api import IncomesApi  # noqa: E501


class TestIncomesApi(unittest.TestCase):
    """IncomesApi unit test stubs"""

    def setUp(self) -> None:
        self.api = IncomesApi()  # noqa: E501

    def tearDown(self) -> None:
        pass

    def test_income_create(self) -> None:
        """Test case for income_create

        """
        pass

    def test_income_destroy(self) -> None:
        """Test case for income_destroy

        """
        pass

    def test_income_list(self) -> None:
        """Test case for income_list

        """
        pass

    def test_income_partial_update(self) -> None:
        """Test case for income_partial_update

        """
        pass

    def test_income_retrieve(self) -> None:
        """Test case for income_retrieve

        """
        pass

    def test_income_update(self) -> None:
        """Test case for income_update

        """
        pass

    def test_incomecategory_create(self) -> None:
        """Test case for incomecategory_create

        """
        pass

    def test_incomecategory_destroy(self) -> None:
        """Test case for incomecategory_destroy

        """
        pass

    def test_incomecategory_list(self) -> None:
        """Test case for incomecategory_list

        """
        pass

    def test_incomecategory_partial_update(self) -> None:
        """Test case for incomecategory_partial_update

        """
        pass

    def test_incomecategory_retrieve(self) -> None:
        """Test case for incomecategory_retrieve

        """
        pass

    def test_incomecategory_update(self) -> None:
        """Test case for incomecategory_update

        """
        pass


if __name__ == '__main__':
    unittest.main()



from rest_framework.serializers import ModelSerializer

from wallet.models.expense import ExpenseCategory, Expense


class ExpenseCategorySerializer(ModelSerializer):
    class Meta:
        model = ExpenseCategory
        fields = "__all__"


class ExpenseSerializer(ModelSerializer):
    class Meta:
        model = Expense
        fields = "__all__"



import os

from celery import Celery
from django.conf import settings


os.environ.setdefault("DJANGO_SETTINGS_MODULE", "moneybox.settings")

app = Celery("moneybox")
app.config_from_object("django.conf:settings", namespace="CELERY")
app.conf.broker_url = settings.CELERY_BROKER_URL
app.autodiscover_tasks()



from django.http import JsonResponse
from django.db import connection
from drf_spectacular.utils import extend_schema

STATUS_OK = {"status": "ok"}
STATUS_ERROR = {"status": "error"}


@extend_schema(tags=["Technical"])
def healthcheck(request):
    try:
        with connection.cursor() as cursor:
            cursor.execute("SELECT 1")
            cursor.fetchone()

        return JsonResponse(STATUS_OK)
    except Exception:
        return JsonResponse(STATUS_ERROR)



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin

from users.models import User


@admin.register(User)
class AbstractUserAdmin(UserAdmin):
    list_display = ("username", "api_user", "email", "first_name", "last_name", "is_staff", "is_active")
    list_filter = ("is_staff", "is_active")
    fieldsets = (
        (None, {"fields": ("username", "password")}),
        ("Personal Info", {"fields": ("first_name", "last_name", "email", "api_user")}),
        ("Permissions", {"fields": ("is_staff", "is_active")}),
        ("Important dates", {"fields": ("last_login", "date_joined")}),
    )
    add_fieldsets = (
        (None, {"classes": ("wide",), "fields": ("username", "password1", "password2", "is_staff", "is_active")}),
    )
    search_fields = ("username", "email", "first_name", "last_name")
    ordering = ("username",)
